# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: locking_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='locking_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15locking_service.proto\x12\x08\x63hip.rpc\"\x07\n\x05\x45mpty\"\x1e\n\x0cLockingState\x12\x0e\n\x06locked\x18\x01 \x01(\x08\x32m\n\x07Locking\x12\x30\n\x03Set\x12\x16.chip.rpc.LockingState\x1a\x0f.chip.rpc.Empty\"\x00\x12\x30\n\x03Get\x12\x0f.chip.rpc.Empty\x1a\x16.chip.rpc.LockingState\"\x00\x62\x06proto3'
)




_EMPTY = _descriptor.Descriptor(
  name='Empty',
  full_name='chip.rpc.Empty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=42,
)


_LOCKINGSTATE = _descriptor.Descriptor(
  name='LockingState',
  full_name='chip.rpc.LockingState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='locked', full_name='chip.rpc.LockingState.locked', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=74,
)

DESCRIPTOR.message_types_by_name['Empty'] = _EMPTY
DESCRIPTOR.message_types_by_name['LockingState'] = _LOCKINGSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Empty = _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), {
  'DESCRIPTOR' : _EMPTY,
  '__module__' : 'locking_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.Empty)
  })
_sym_db.RegisterMessage(Empty)

LockingState = _reflection.GeneratedProtocolMessageType('LockingState', (_message.Message,), {
  'DESCRIPTOR' : _LOCKINGSTATE,
  '__module__' : 'locking_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.LockingState)
  })
_sym_db.RegisterMessage(LockingState)



_LOCKING = _descriptor.ServiceDescriptor(
  name='Locking',
  full_name='chip.rpc.Locking',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=76,
  serialized_end=185,
  methods=[
  _descriptor.MethodDescriptor(
    name='Set',
    full_name='chip.rpc.Locking.Set',
    index=0,
    containing_service=None,
    input_type=_LOCKINGSTATE,
    output_type=_EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Get',
    full_name='chip.rpc.Locking.Get',
    index=1,
    containing_service=None,
    input_type=_EMPTY,
    output_type=_LOCKINGSTATE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LOCKING)

DESCRIPTOR.services_by_name['Locking'] = _LOCKING

# @@protoc_insertion_point(module_scope)
